#include "atari.h"
#include "config.h"
#include "monitor.h"
#include "pokeysnd.h"
#include "sound.h"

#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "atari_gp32.h"

#include "gp2xminlib.h"

// We have dual sound buffers
// i) Controlled by gpsoundbuffer.c. Callback when we need more data to play.
// ii) Controlled by emulator. Filled every frame.
//
// Strategy: Fill emulator buffer in main code. Copy into play buffer on callback.
//
// Reason: Emulator does not like callbacks on interrupt into the sound code.
// Mostly works, but volume only seems broken among other things.
// Quite interesting really since the SDL port appears to use the callback method only.
// Another reason: Decouples emulator time from GP32 time

int soundBufferSamplesStored = 0;

#define DEFDSPRATE 22050
static int soundPlaying = FALSE;

#ifdef STEREO
	#define SNDBUFSIZE 10000
#else
	#define SNDBUFSIZE 5000
#endif
static char * atariSoundBuffer = 0;
static int atariSoundBufferStart = 0;
static int atariSoundBufferEnd = 0;

static pthread_t soundThread;
static pthread_mutex_t soundMutex = PTHREAD_MUTEX_INITIALIZER;

//////////////////////////////////////////////////////////////////////////
// Forward declarations of some internal functions
static int soundBufferRollingDistance(int start, int end, int length);
static void * soundThreadCode(void *);

//////////////////////////////////////////////////////////////////////////
// Functions called externally by the emulator code...
void Sound_Initialise(int *argc, char *argv[])
{
	// And sound
#ifdef STEREO
	stereo_enabled = 0;
#endif

	atariSoundBuffer = (char *)malloc(SNDBUFSIZE);
#ifdef STEREO
        //GpPcmInit(PCM_S22, PCM_8BIT);
	Pokey_sound_init(FREQ_17_EXACT, DEFDSPRATE, 2);
	gp2x_sound_init(22050,8,1);
#else
        //GpPcmInit(PCM_M22, PCM_8BIT);
	Pokey_sound_init(FREQ_17_EXACT, DEFDSPRATE, 1);
	gp2x_sound_init(22050,8,0);
#endif
	pthread_attr_t attr;
	struct sched_param param;
	param.sched_priority = 80;
	pthread_attr_init(&attr);
	//pthread_attr_setschedpolicy(&attr,SCHED_RR);
	//pthread_attr_setschedparam(&attr, &param);
	nice(-20);

	pthread_create(&soundThread, &attr, soundThreadCode, 0);
}

void Sound_Pause(void)
{
	if (!soundPlaying)
		return;

	soundPlaying = FALSE;
}

void Sound_Continue(void)
{
	if (soundPlaying)
		return;

	soundPlaying = TRUE;	
}

void Sound_Exit(void)
{
	Sound_Pause();
}

void Sound_Update()
{
	if (soundPlaying)
	{
		pthread_mutex_lock(&soundMutex);
			
		// Fill buffer
		int samplesToGet = 0;
		int samplesLeft = 0;
		samplesLeft = soundBufferRollingDistance(atariSoundBufferStart, atariSoundBufferEnd, SNDBUFSIZE);

#ifdef STEREO
		samplesToGet = 22050*deltatime*2; // 22050/50 *2 approx
#else
		samplesToGet = 22050*deltatime;
#endif

		if (samplesLeft < samplesToGet)
		{
			samplesToGet *= 1.1;
		}
		else if (samplesLeft > (SNDBUFSIZE-samplesToGet))
		{
			samplesToGet *= 0.9;
		}
		if (samplesToGet&1)
			samplesToGet-=1;

		if (samplesToGet > 0)
		{
			if (SNDBUFSIZE < atariSoundBufferEnd + samplesToGet) //overflow
			{
				int remaining = SNDBUFSIZE - atariSoundBufferEnd;
				Pokey_process(atariSoundBuffer + atariSoundBufferEnd, remaining);
				Pokey_process(atariSoundBuffer, samplesToGet-remaining);
				atariSoundBufferEnd = samplesToGet-remaining;
			}
			else
			{
				Pokey_process(atariSoundBuffer + atariSoundBufferEnd, samplesToGet);
				atariSoundBufferEnd += samplesToGet;
			}
		}

		pthread_mutex_unlock(&soundMutex);
	}
}

//////////////////////////////////////////////////////////////////////////
// Helper function for sound circular buffer
static int soundBufferRollingDistance(int start, int end, int length)
{
	if (end > start)
		return (end-start);
	if (end < start)
		return (length-start + end);

	return 0;
}

// Thread, which calls "interrupt driven callback"
static void * soundThreadCode (void * param)
{
	while (1)
	{
		gp2x_sound_play();
	}
	return 0;
}

// Interrupt driven callback from the sound buffer code - well not quite...
int gp2x_sound_frame(void * userdata, void * streamIn, int lenIn)
{
	pthread_mutex_lock(&soundMutex);

	int success = 0;
	
	static int last = 0;
	unsigned char * stream = (unsigned char *) streamIn;
	int len=lenIn;
	
	int samplesAvailable = soundBufferRollingDistance(atariSoundBufferStart,atariSoundBufferEnd, SNDBUFSIZE);
	if (samplesAvailable>len)
	{
		if ((atariSoundBufferStart+len) < SNDBUFSIZE)
		{
			memcpy(stream, atariSoundBuffer+atariSoundBufferStart, len);
			atariSoundBufferStart+=len;
		}
		else
		{
			int remaining = SNDBUFSIZE - atariSoundBufferStart;
			memcpy(stream, atariSoundBuffer+atariSoundBufferStart, remaining);

			int fromStart = len-remaining;
			if (fromStart)
			{
				memcpy(stream+remaining, atariSoundBuffer, fromStart);
			}
			atariSoundBufferStart = fromStart;
		}
//	}
//	else
//	{
//		memset(stream, last, len);
//	}

		last = stream[len-1];

		// Store count of samples -> for external use in display status text
		soundBufferSamplesStored = samplesAvailable;

		success = 1;
	}

	pthread_mutex_unlock(&soundMutex);
	
//	debugText = debugTextBuffer;
//	sprintf(debugTextBuffer,"%4d           ", soundBufferSamplesStored);
	
//	if (soundBufferSamplesStored < 500)
//		sprintf(debugTextBuffer+6, "ALERT");
//
	return success;
}

