#include "atari.h"
#include "config.h"
#include "monitor.h"
#include "colours.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "atari_gp32.h"
#include "colour_hsv.h"

//#define COUNT_FPS
#ifdef COUNT_FPS
#define DEBUG_TEXT
#endif
//#define DEBUG_TEXT

// Used externally as an easy way of getting text on the screen!
char * debugTextBuffer = 0;
char * debugText =0;

// Brightness (used externally in state save)
char brightness = 2;

// Screen buffers -> theoretically we double buffer.
// In practice it is not bad without so it is off ATM
// I have performance concerns with GpSwapBuffer
// Need to investigate further
// Update: Just tried -> performance reduced by approx 30-40%
// WTF? Guess it must be waiting for sync?
static GPDRAWSURFACE gpDraw[2];
static int backbuf = 0;
#define DBL_BUFFER
#define DBL_BUFFER_LCD
// << Define to change LCD regs directly for double buffering
// This (mostly) does not wait for line (note that LCD does it itself because of DMA...)

// Interrupt driven frame count
// Every second we get called back and store the value here
#ifdef COUNT_FPS
static char countFPSResult[256];
static int drawFPS = 0;
static int realFPS = 0;
static int realFPSCount = 0;
static int drawFPSCount = 0;
void fpsCallback() { drawFPS = drawFPSCount; drawFPSCount = 0; realFPS = realFPSCount; realFPSCount = 0; }
void countFPS(int fullRender);
#endif

// Notes on DIRTYRECT
// Tried it and found it to be slower than without.
// Seemed to be that when it was enabled the antic rendering took 10ms longer. Since the copy to screen takes 5ms that isn't a gain.
#ifdef DIRTYRECT
static char all_dirty = 0;
void entire_screen_dirty() { all_dirty = 1; }
#endif

#ifndef PI
#define PI 3.142
#endif

///////////////////////////////////////////////////////////////
// Interface, called externally
GPDRAWSURFACE * Atari_DisplayActiveBuffer()
{
	return &gpDraw[backbuf];
}

void Atari_DisplayInitPhase2()
{
	GpGraphicModeSet(8, NULL);
	GpLcdSurfaceGet(&gpDraw[0], 0);
	GpLcdSurfaceGet(&gpDraw[1], 1);

	GpRectFill(NULL, &gpDraw[0], 0,0,320,240, 0x00);
	GpSurfaceSet(&gpDraw[0]);
}

void Atari_DisplayInitialise()
{
#ifdef DEBUG_TEXT
	debugTextBuffer = (char *) malloc(256);
#endif

	// Set palette (colortable is setup by emulator palette code)
	Atari_DisplayBrightness(brightness);

#ifdef COUNT_FPS
	// FPS count
	GpTimerOptSet(gpNextTimer, 1, 0, fpsCallback);
	GpTimerSet(gpNextTimer++);
#endif
}

void Atari_DisplayBrightness(int level)
{
	GP_HPALETTE h_pal;
	GP_PALETTEENTRY entry[256];
	int i;

	brightness = level;
	int sinScaleLevel = level&7;
	int fluLevel = level&8;

	for( i = 0 ; i < 256 ; i++ )
	{
		entry[i] = 0;

		int R = (colortable[i]&0xff0000) >> 16;
		int G = (colortable[i]&0xff00) >> 8;
		int B = (colortable[i]&0xff) >> 0;

		double sinScale = 256;
		switch (sinScaleLevel)
		{
		case 1:
			sinScale = 256;
			break;
		case 2:
			sinScale = 320;
			break;
		case 3:
			sinScale = 380;
			break;
		case 4:
			sinScale = 500;
			break;
		}

		double Rd=R,Gd=G,Bd=B;

		if (fluLevel)
		{
			double H,S,V;
			rgb2hsv(Rd,Gd,Bd,&H,&S,&V);
			double Vold = V;
			if (Vold < 0.5)
			{
				V*=1.1;
			}
			if (Vold < 0.3)
			{
				V*=1.1;
			}
			if (Vold < 0.1)
			{
				V*=1.1;
			}
			hsv2rgb(H,S,V,&Rd,&Gd,&Bd);
		} 
		// FLU can be a bit washed out -> increase the saturation

		if (sinScaleLevel != 0)
		{
			Rd = PI*(double)Rd/(double)512;
			Gd = PI*(double)Gd/(double)512;
			Bd = PI*(double)Bd/(double)512;

			Rd = sin(Rd)*sinScale;
			Gd = sin(Gd)*sinScale;
			Bd = sin(Bd)*sinScale;
		}
		// Since GP32 is so fscking dull, need to boost mid range
		// Don't really want to boost dark stuff since I like true black
		// So use a nice curve

		if (fluLevel || sinScaleLevel)
		{
			R = Rd<256 ? Rd : 255;
			G = Gd<256 ? Gd : 255;
			B = Bd<256 ? Bd : 255;
		}

		entry[i] |= (R&0xf8) << (8);
		entry[i] |= (G&0xf8) << (3);
		entry[i] |= (B&0xf8) >> (2);
	}

	h_pal = GpPaletteCreate( 256, entry );
	GpPaletteSelect( h_pal );
	GpPaletteRealize();
}

void Atari_DisplayScreen(UBYTE * screen)
{	
	int i,j;

	// Passed a null pointer if we are not expected to draw
	// Called every frame regardless...
#ifdef DEBUG_TEXT
	int oldbackbuf = backbuf;
#endif

	if (screen)
	{
		int width = (320 - menuWidth)/4;

		int offset = menuWidth + abs((ATARI_WIDTH-320)/2);

		char * destOrigin = (240*menuWidth) + gpDraw[backbuf].ptbuffer;
		destOrigin += 240;
		char * dest;

		int srcOffset = -(4*width) + (ATARI_WIDTH*4);
		char * src1 = (char *) (screen + offset);
		char * src2 = src1 + ATARI_WIDTH;
		char * src3 = src2 + ATARI_WIDTH;
		char * src4 = src3 + ATARI_WIDTH;

		int height = (240 - menuHeight)/4;

#ifdef DIRTYRECT
		if (all_dirty)
		{
			all_dirty = 0;
#endif
			// OPTIMISED VERSION THAT DOESN'T CHECK DIRTY RECT!
			// Seems to be faster than the asm XlatVGABuffer :-)
			// Tried writing this one in ASM and with the LW read/writes only (no logic) it took the same time	
			// So guess we're memory bound here (not surprising)
			// Wonder if there are any bus setting tweaks that would help...
			for (i=0; i<height; ++i)
			{
				destOrigin-=4;
				dest = destOrigin;

				for (j=0; j<width; ++j)
				{
					int value1 = *(int *)src1;
					src1+=4;
					int value2 = *(int *)src2;
					src2+=4;
					int value3 = *(int *)src3;
					src3+=4;
					int value4 = *(int *)src4;
					src4+=4;

					*(int *)dest = ((value1&0xff)<<24) | ((value2&0xff)<<16) | ((value3&0xff)<<8) | ((value4&0xff)<<0);
					dest+=240;

					*(int *)dest = ((value1&0xff00)<<16) | ((value2&0xff00)<<8) | ((value3&0xff00)<<0) | ((value4&0xff00)>>8);
					dest+=240;

					*(int *)dest = ((value1&0xff0000)<<8) | ((value2&0xff0000)<<0) | ((value3&0xff0000)>>8) | ((value4&0xff0000)>>16);
					dest+=240;
		
					*(int *)dest = ((value1&0xff000000)<<0) | ((value2&0xff000000)>>8) | ((value3&0xff000000)>>16) | ((value4&0xff000000)>>24);
					dest+=240;
				}

				src1 += srcOffset;
				src2 += srcOffset;
				src3 += srcOffset;
				src4 += srcOffset;
			}
#ifdef DIRTYRECT
		}
		else
		{
			// USE DIRTY RECT!
			short *dirty1 = screen_dirty + 4; //(offset/16);
			short *dirty2 = dirty1 + (ATARI_WIDTH/16); //8*short size
			short *dirty3 = dirty2 + (ATARI_WIDTH/16);
			short *dirty4 = dirty3 + (ATARI_WIDTH/16);
			short dirtyIter; //Each dirty SW contains 16 pixels -> so 4 iterations...
			short dirtyComp = 0;
			short dirtyVal1;
			short dirtyVal2;
			short dirtyVal3;
			short dirtyVal4;
			short dirtyOffset = -(width/4) + ((ATARI_WIDTH/16)*4);

			for (i=0; i<60; ++i)
			{
				destOrigin-=4;
				dest = destOrigin;

				dirtyIter = 12;
				for (j=0; j<width; ++j)
				{
					dirtyIter += 4;
					if (dirtyIter == 16)
					{
						dirtyVal1 = *dirty1++;
						dirtyVal2 = *dirty2++;
						dirtyVal3 = *dirty3++;
						dirtyVal4 = *dirty4++;

						dirtyIter = 0;
					}
					dirtyComp = dirtyIter&0xf8;
					char blockDirty = 0;
					blockDirty |= (dirtyVal1>>dirtyComp)&1;
					blockDirty |= (dirtyVal2>>dirtyComp)&1;
					blockDirty |= (dirtyVal3>>dirtyComp)&1;
					blockDirty |= (dirtyVal4>>dirtyComp)&1;
			
					if (blockDirty)
					{
						int value1 = *(int *)src1;
						src1+=4;
						int value2 = *(int *)src2;
						src2+=4;
						int value3 = *(int *)src3;
						src3+=4;
						int value4 = *(int *)src4;
						src4+=4;

						*(int *)dest = ((value1&0xff)<<24) | ((value2&0xff)<<16) | ((value3&0xff)<<8) | ((value4&0xff)<<0);
						dest+=240;

						*(int *)dest = ((value1&0xff00)<<16) | ((value2&0xff00)<<8) | ((value3&0xff00)<<0) | ((value4&0xff00)>>8);
						dest+=240;

						*(int *)dest = ((value1&0xff0000)<<8) | ((value2&0xff0000)<<0) | ((value3&0xff0000)>>8) | ((value4&0xff0000)>>16);
						dest+=240;
			
						*(int *)dest = ((value1&0xff000000)<<0) | ((value2&0xff000000)>>8) | ((value3&0xff000000)>>16) | ((value4&0xff000000)>>24);
						dest+=240;
					}
					else
					{
						src1+=4;
						src2+=4;
						src3+=4;
						src4+=4;
						dest+=960;
					}
				}

				src1 += srcOffset;
				src2 += srcOffset;
				src3 += srcOffset;
				src4 += srcOffset;

				dirty1 += dirtyOffset;
				dirty2 += dirtyOffset;
				dirty3 += dirtyOffset;
				dirty4 += dirtyOffset;
			}
		}

		// Clear dirty rect
		memset(screen_dirty, 0, ATARI_WIDTH*ATARI_HEIGHT/8); 
#endif

#ifdef DBL_BUFFER
#ifdef DBL_BUFFER_LCD
		static int bb1A=0, bb1B=0, bb2A=0, bb2B=0;
		static volatile int * LCDADDR1 = (int *)0x14a00014;
		static volatile int * LCDADDR2 = (int *)0x14a00018;
		static volatile int * LCDCON1  = (int *)0x14a00000;

		if (!bb1A || !bb1B || !bb2A || !bb2B)
		{
			GpSurfaceFlip(&gpDraw[backbuf]);
	
			if (backbuf)
			{
				bb1A = *LCDADDR1;
				bb1B = *LCDADDR2;
			}
			else
			{
				bb2A = *LCDADDR1;
				bb2B = *LCDADDR2;
			}
		}
		else
		{
			while (((*LCDCON1 >> 18)&0x1ff) == 1);
			while (((*LCDCON1 >> 18)&0x1ff) == 0);
			if (backbuf)
			{
				*LCDADDR1 = bb1A;
				*LCDADDR2 = bb1B;
			}
			else
			{
				*LCDADDR1 = bb2A;
				*LCDADDR2 = bb2B;
			}
		}
#else
		GpSurfaceFlip(&gpDraw[backbuf]);
#endif

		backbuf^=1;
#endif
	}

//	debugText = debugTextBuffer;
//	sprintf(debugTextBuffer, "RTC:%8d %8d\n", rtcTick, GpTickCountGet());

#ifdef DEBUG_TEXT
	if (debugText)
	{
		GpRectFill(NULL, &gpDraw[oldbackbuf],5,220,305,10, 0x00);
		GpTextOut(0, &gpDraw[oldbackbuf], 9, 220, debugText, 0xff);
	}
#endif

#ifdef COUNT_FPS
	countFPS(screen!=0);
	GpRectFill(NULL, &gpDraw[oldbackbuf],5,230,305,10, 0x00);
	GpTextOut(0, &gpDraw[oldbackbuf], 9, 230, countFPSResult, 0xff);
#endif
}

///////////////////////////////////////////////////////////////
// Internal
#ifdef COUNT_FPS
void countFPS(int fullRender)
{
	static int first = 1;
	static int lastTicks = 0;
	static int totalTicks = 0;
	static int totalFrames = 0;
	static int skipCount = 0;

	static int fullRenderTicks[10];
	static int fullRenderPos = 0;

	if (first)
	{
		lastTicks = GpTickCountGet();
		first = 0;
	}
	else
	{
		int now = GpTickCountGet();
		int taken = now - lastTicks;
		int avgFrames = ((++totalFrames)*1000)/totalTicks;

		lastTicks = now;
		totalTicks += taken;

		if (fullRender) // Did we render this frame?
		{
			fullRenderPos++;
			if (fullRenderPos == 10)
				fullRenderPos = 0;

			fullRenderTicks[fullRenderPos] = taken;

			drawFPSCount++;
		}
		else
			skipCount++;
		realFPSCount++;

		int fullRenderTickCount=0;
		int k;
		for (k=0;k<10;++k)
			fullRenderTickCount += fullRenderTicks[k];

		sprintf(countFPSResult, "dps:%2d eps:%2d skp:%d rms:%2d rfp:%2d", drawFPS, realFPS, skipCount, fullRenderTickCount/10, 10000/fullRenderTickCount);

	}
}
#endif

